from .cli.imports import *

class BedRecord:
    def __init__(self, chrom, chromStart, chromEnd, name=None, score=None, strand=None):
        self.chrom = chrom
        self.chromStart = int(chromStart)
        self.chromEnd = int(chromEnd)
        self.name = name
        self.score = score
        self.strand = strand

    def __repr__(self):
        return f"BedRecord(chrom={self.chrom}, chromStart={self.chromStart}, chromEnd={self.chromEnd}, " \
               f"name={self.name}, score={self.score}, strand={self.strand})"

def parse_bed_file(bed_file_path):
    records = []
    
    with open(bed_file_path, 'r') as file:
        for line in file:
            # Skip empty lines or comments
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            
            fields = line.split('\t')
            
            # BED file typically has at least 3 columns, handle additional ones if present
            chrom = fields[0]
            chromStart = fields[1]
            chromEnd = fields[2]
            
            # Optional fields
            name = fields[3] if len(fields) > 3 else None
            score = fields[4] if len(fields) > 4 else None
            strand = fields[5] if len(fields) > 5 else None

            # Create a BedRecord object and add it to the list
            record = BedRecord(chrom, chromStart, chromEnd, name, score, strand)
            records.append(record)
    
    return records

def parse_bed2d_file(bed_file_path):
    records = []
    
    with open(bed_file_path, 'r') as file:
        for line in file:
            # Skip empty lines or comments
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            
            fields = line.split('\t')
            
            # BED2D file typically has at least 3 columns, handle additional ones if present
            chrom = fields[0]
            chromStart = fields[1]
            chromEnd = fields[2]
            name = fields[3] if len(fields) > 6 else None
            score = fields[4] if len(fields) > 8 else None
            strand = fields[5] if len(fields) > 10 else None
            k = len(fields) // 2 - 1 
            record = BedRecord(chrom, chromStart, chromEnd, name, score, strand)

            chrom2 = fields[k + 1]
            chromStart2 = fields[k + 2]
            chromEnd2 = fields[k + 3]
            name2 = fields[k + 4] if len(fields) > 6 else None
            score2 = fields[k + 5] if len(fields) > 8 else None
            strand2 = fields[k + 6] if len(fields) > 10 else None
            record2 = BedRecord(chrom2, chromStart2, chromEnd2, name2, score2, strand2)
            
            
            # Create a BedRecord object and add it to the list
            records.append(tuple((record, record2)))
    
    return records

def parse_gff(in_file):
    """Parses gff file into python dataframe."""
    gff = GFF.parse(in_file)
    gff_df = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand"])
    
    for rec in gff:
        for feature in rec.features:
                loc = feature.location
                gff_tmp = {
                    "Name": feature.qualifiers['Name'][0] if "Name" in feature.qualifiers else feature.id,
                    "Chromosome": rec.id,
                    "Start": loc.start,
                    "End": loc.end,
                    "Strand": loc.strand,
                }
                gff_df = gff_df._append(gff_tmp, ignore_index=True)
    return gff_df

def parse_bed(in_file, default_strand=1, default_name="Locus"):
    """Parses bed file into python dataframe."""
    bed = parse_bed_file(in_file)
    bed_df = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand"])
    
    i = 0
    for rec in bed:
        bed_tmp = {
                    "Name": rec.name if rec.name != None else f"{default_name}_{i}",
                    "Chromosome": rec.chrom,
                    "Start": rec.chromStart,
                    "End": rec.chromEnd,
                    "Strand": 1 if rec.strand == '+' else -1 if rec.strand == '-' else default_strand
                }
        bed_df = bed_df._append(bed_tmp, ignore_index=True)
        if rec.name == None:
            i += 1
    return bed_df

def parse_bed2d(in_file, default_strand=1, default_name="Pair"):
    """Parses bed file into python dataframe."""
    bed2d = parse_bed2d_file(in_file)
    bed2d_df = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand", "Name2", "Chromosome2", "Start2", "End2", "Strand2"])
    
    i = 0
    for rec1, rec2 in bed2d:
        bed2d_tmp = {
                    "Name": rec1.name if rec1.name != None else f"{default_name}_{i}_loc1",
                    "Chromosome": rec1.chrom,
                    "Start": rec1.chromStart,
                    "End": rec1.chromEnd,
                    "Strand": 1 if rec1.strand == '+' else -1 if rec1.strand == '-' else default_strand,
                    "Name2": rec2.name if rec2.name != None else f"{default_name}_{i}_loc2",
                    "Chromosome2": rec2.chrom,
                    "Start2": rec2.chromStart,
                    "End2": rec2.chromEnd,
                    "Strand2": 1 if rec2.strand == '+' else -1 if rec2.strand == '-' else default_strand
                }
        bed2d_df = bed2d_df._append(bed2d_tmp, ignore_index=True)
        if rec1.name == None or rec2.name == None:
            i += 1
    return bed2d_df

def genes_in_rec(bed_rec, gff, overlap="flex"):
    """Returns the records from gff included in the bed record bed_rec interval."""
    pos1_rec = min(bed_rec.chromStart, bed_rec.chromEnd)
    pos2_rec = max(bed_rec.chromStart, bed_rec.chromEnd)

    genes = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand"])
    for _, position in gff.iterrows():
        if position["Chromosome"] != bed_rec.chrom:
            continue

        pos1 = min(position["Start"], position["End"])
        pos2 = max(position["Start"], position["End"])
        match overlap:
            case "flex":
                # included if the position overlaps the regions, even not completely
                if (pos1 >= pos1_rec and pos1 <= pos2_rec) or (pos2 >= pos1_rec and pos2 <= pos2_rec):
                    genes = genes._append(position, ignore_index = True)
            case "strict":
                # included only if the position is completely in the region
                if pos1 >= pos1_rec and pos2 <= pos2_rec:
                    genes = genes._append(position, ignore_index = True)

    return genes

def parse_bed_annotated(bed_file, gff_file, overlap="flex"):
    """Annotates each interval in the provided bed file with the found sequences in the gff file."""
    bed = parse_bed_file(bed_file)
    gff = parse_gff(gff_file)
    genes_df = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand"])
    for rec in bed:
        rec_genes = genes_in_rec(rec, gff, overlap=overlap)
        genes_df = pd.concat([genes_df, rec_genes])
    return genes_df

def parse_tracks(tracks, threshold = None, percentage = None, gff=None, default_strand=1, binning=1000):
    """Parses a track file into a positions table. Applies threshold AND percentage if provided. If a gff file has been provided, will compute average track value per record."""
    bigwig_file = pyBigWig.open(tracks)

    percentage_threshold = None
    if threshold != None:
        min_max, threshold_value = threshold

    if gff != None:
        positions = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand", "Track"])
        for rec in GFF.parse(gff):
            for feature in rec.features:
                    loc = feature.location
                    value = bigwig_file.stats(rec.id, loc.start, loc.end)[0]

                    if threshold != None and value != None:
                        match min_max:
                            case 'min':
                                value = value if value >= threshold_value else None
                            case 'max':
                                value = value if value <= threshold_value else None

                    if value != None:
                        pos_tmp = {
                            "Name": feature.qualifiers['Name'][0] if "Name" in feature.qualifiers else feature.id,
                            "Chromosome": rec.id,
                            "Start": loc.start,
                            "End": loc.end,
                            "Strand": loc.strand,
                            "Track": value
                        }
                        positions = positions._append(pos_tmp,  ignore_index = True)
        # as it requires all gff sequences annotations with tracks, percentage is applied after said annotation
        if percentage != None:
            app_end, quantile = percentage
            match app_end:
                case 'high':
                    quantile = 1 - quantile/100
                case 'low':
                    quantile = quantile/100
            percentage_threshold = np.nanquantile(list(positions['Track']), quantile)
            positions = positions[positions['Track'] >= percentage_threshold].reset_index(drop=True)

    else:
        if percentage != None:
            app_end, quantile = percentage
            match app_end:
                case 'high':
                    quantile = 1 - quantile/100
                case 'low':
                    quantile = quantile/100
            all_values = np.concatenate([bigwig_file.values(chrom, 0, bigwig_file.chroms(chrom)) for chrom in bigwig_file.chroms().keys()])
            percentage_threshold = np.nanquantile(all_values, quantile)

        positions = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand", "Track"])
        k = 1
        for chrom in bigwig_file.chroms().keys():
            for i in range(0, bigwig_file.chroms(chrom) - binning, binning):
                start = i
                stop = i + binning
                value = bigwig_file.stats(chrom, start, stop)[0]

                if threshold != None and value != None:
                    match min_max:
                        case 'min':
                            value = value if value >= threshold_value else None
                        case 'max':
                            value = value if value <= threshold_value else None

                if percentage_threshold != None and value != None:
                    match app_end:
                        case 'high':
                            value = value if value >= percentage_threshold else None
                        case 'low':
                            value = value if value <= percentage_threshold else None
                if value != None:
                    pos_tmp = {
                        "Name": f"Selection_{k}",
                        "Chromosome": chrom,
                        "Start": start,
                        "End": stop,
                        "Strand": default_strand,
                        "Track": value
                    }
                    positions = positions._append(pos_tmp, ignore_index = True)
                    k += 1

    return positions, bigwig_file



        




import click

import hicue.hicue as h

@click.command("annotate")
@click.argument('gff_file', type=click.Path(exists=True, dir_okay=False))
@click.argument('bed_file', type=click.Path(exists=True, dir_okay=False))
@click.option('--overlap', type=click.Choice(['strict', 'flex'], case_sensitive=False), default='flex', help="When evaluating the belonging of a position to an interval, sets the severity of the discrimination: 'strict' will only allow position having start and end positions within the interval, whereas 'flex' will consider the position even if the overlap is not complete. Default value: strict.")
def annotate(gff_file, bed_file):
    h.annotate(gff_file, bed_file)
    



import click

from .imports import *

from hicue.parser import *

# accepts a list of cool files or a file containing a list of cool, one per line
class coolType(click.ParamType):
    name="cool"

    def is_cool(self, file):
        try:
            return cooler.fileops.is_cooler(file)
        except:
            return False

    def convert(self, value, param, ctx):
        if self.is_cool(value):
            return [value]
        if not os.path.isfile(value):
            files = value.split(',')
            for file in files:
                if not self.is_cool(file):
                    self.fail(f"{file} is not a valid cool file", param, ctx)
            return files
        else:
            files = []
            with open(value, 'r') as f:
                for line in f.readlines():
                    file = line.replace('\n', '').replace(' ', '')
                    if not self.is_cool(file):
                        self.fail(f"{file} is not a valid cool file", param, ctx)
                    else:
                        files.append(file)
            return files

class IntListType(click.ParamType):
    name="int_list"

    def convert(self, value, param, ctx):
        list = value.split(',')
        ints = []
        for i in list:
            if not isinstance(int(i), int):
                self.fail(f"{value} is not a comma-separated integer list.")
            if int(i) < 0:
                self.fail(f"All integers provided in {param.name} must be positive integers.")
            ints.append(int(i))
        return ints
    
class StrListType(click.ParamType):
    name="str_list"

    def convert(self, value, param, ctx):
        strs = value.split(',')
        for s in strs:
            if len(s) == 0:
                self.fail(f"Empty strings. Comma must separate two distinct values; if a single string is passed, no comma must be found in {param.name}.")
        return strs

class PositionFileType(click.ParamType):
    name="position file"

    #  accepts bed and gff files. Returns a tuple indicating the type and path of the file
    def convert(self, value, param, ctx):
        if not os.path.isfile(value):
            self.fail(f"{value} is not an existing file.")
        try:
            in_handle = open(value)
            recs = GFF.parse(in_handle)
            if len(list(recs)) == 0:
                raise Exception
            return 'gff', value
        except:
            try:
                bed = parse_bed_file(value)
                return 'bed', value
            except Exception as e:
                self.fail(f"{value} is not an valid position file (gff or bed format expected) : {e}")

class Position2dFileType(click.ParamType):
    name="position file"

    #  accepts bed and gff files. Returns a tuple indicating the type and path of the file
    def convert(self, value, param, ctx):
        if not os.path.isfile(value):
            self.fail(f"{value} is not an existing file.")
        try:
            parse_bed2d_file(value)
            return value
        except Exception as e:
            self.fail(f"{value} is not an valid 2d position file (bed2d format expected) : {e}")

class GffFileType(click.ParamType):
    name="gff file"

    #  accepts bed and gff files. Returns a tuple indicating the type and path of the file
    def convert(self, value, param, ctx):
        if not os.path.isfile(value):
            self.fail(f"{value} is not an existing file.")
        try:
            in_handle = open(value)
            recs = GFF.parse(in_handle)
            if len(list(recs)) == 0:
                raise Exception
            return value
        except Exception as e:
            self.fail(f"{value} is not an valid gff file : {e}")

class TrackFileType(click.ParamType):
    name="track file"

    #  accepts bw files
    def convert(self, value, param, ctx):
        try:
            bw = pyBigWig.open(value)
            if bw.isBigWig():
                return value
            else:
                self.fail(f"{value} provided for {param.name} is not in the bigWig format.")
        except:
            self.fail(f"{value} provided for {param.name} is not in the bigWig format.")



COOL = coolType()
INT_LIST = IntListType()
STR_LIST = StrListType()
POSITION_FILE = PositionFileType()
POSITION2D_FILE = Position2dFileType()
TRACK_FILE = TrackFileType()
GFF_FILE = GffFileType()


import click

from . import (
    extract,
    tracks,
    extract2d, 
    # separate,
    # annotate
)

@click.group()
@click.pass_context
def cli(ctx):
    pass

cli.add_command(extract.extract)
cli.add_command(tracks.tracks)
cli.add_command(extract2d.extract2d)
# cli.add_command(separate.separate)
# cli.add_command(annotate.annotate)

if __name__ == "__main__":
    cli()


from .utils import *
from .displays import *
from .parser import *

def extract(cool_files, positions, outpath, params):

    if not os.path.exists(outpath):
        os.mkdir(outpath)

    # parsing parameters
    gff = params['gff']
    windows = params['windows']
    detrending = params['detrending']
    nb_pos = params['nb_pos']
    max_dist = params['random_max_dist']
    loops = params['loops']
    min_dist = params['min_dist']
    diagonal_mask = params['diagonal_mask']
    trans_contact = params['trans_contact']
    circular_chromosomes = params['circular_chromosomes']
    display_strand = params['display_strand']
    output_format = params['output_formats']
    compute_pileup = params['pileup']
    plot_loci = params['loci']
    method = params['method']
    flip = params['flip']
    cmap = params['cmap_pileup']
    display_sense = params['display_sense']
    center = params["center"]
    separate_by = params["separate_by"]
    separate_regions = params["separation_regions"]
    overlap = params["overlap"]
    contact_separation = params["contact_separation"]
    contact_range = params["contact_range"]
    ps_detrending = params["detrending"] == "ps"

    # parsing gff file
    pos_type, pos_file = positions
    match pos_type:
        case 'gff':
            positions_parsed = parse_gff(pos_file) 
        case 'bed':
            # if gff file is provided, annotating each line of the bed file with its genes. The submatrices will be computed on such positions.
            if gff != None:
                positions_parsed = parse_bed_annotated(pos_file, gff, overlap=overlap)
                # TODO write position file
                # write_positions(positions_parsed, outpath)
            else:
                positions_parsed = parse_bed(pos_file, default_strand=1)

    data_title = pos_file.split('/')[-1].split('.')[0].replace('/', '_')

    selected_positions = separate_positions(positions_parsed, data_title, separate_by=separate_by, separate_regions=separate_regions, overlap=overlap, outpath=outpath)
    random_locus = []
    for position_name, positions in selected_positions.items():
        for cool_path in cool_files:
            cool = cooler.Cooler(cool_path)
            bins = cool.binsize
            if detrending == "patch":
                random_locus = get_random_from_locus(cool, positions_parsed, nb_pos=nb_pos, max_dist=max_dist) if len(random_locus) == 0 else random_locus

            matrix_outfolder = f"{outpath}/{cool.filename.split('/')[-1].split('.')[0]}"
            if not os.path.exists(matrix_outfolder):
                os.mkdir(matrix_outfolder)


            for window in windows:
                outfolder = f"{matrix_outfolder}/window_{window//1000}kb"
                if not os.path.exists(outfolder):
                    os.mkdir(outfolder)

                # retrieving submatrices
                submatrices = compute_submatrices(cool, 
                                                  position_name, 
                                                  positions, 
                                                  bins, 
                                                  window, 
                                                  loops=loops, 
                                                  min_dist=min_dist,
                                                  circular=circular_chromosomes, 
                                                  trans_contact=trans_contact, 
                                                  diagonal_mask=diagonal_mask, 
                                                  center=center, 
                                                  sort_contact=contact_separation, 
                                                  contact_range = contact_range,
                                                  ps_detrend = ps_detrending)

                if detrending == "patch":
                    random_submatrices = compute_submatrices(cool, 
                                                            position_name, 
                                                            random_locus, 
                                                            bins, 
                                                            window, 
                                                            loops=loops, 
                                                            min_dist=min_dist,
                                                            circular=circular_chromosomes, 
                                                            trans_contact=False, 
                                                            center = center, 
                                                            sort_contact=contact_separation, 
                                                            contact_range = contact_range)
                 
                for name in submatrices.keys():
                    if len(submatrices[name]) == 0:
                        continue
                    # displaying submatrices
                    if plot_loci:
                        display_submatrices(submatrices[name], positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense)
                    display_all_submatrices(submatrices[name], positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense)
                    
                    # computing pileup
                    if compute_pileup:
                        ## aggregating the matrices
                        pileup_matrices = get_windows(submatrices[name], positions, flip)
                        match method:
                            case "median":
                                pileup = np.apply_along_axis(np.nanmedian, 0, pileup_matrices)
                            case "mean":
                                pileup = np.apply_along_axis(np.nanmean, 0, pileup_matrices)

                        ## detrending
                        if name[-len("trans"):] != "trans":
                            match detrending:
                                case "patch":
                                    if nb_pos >= 1:
                                        random_pileup_matrices = get_windows(random_submatrices[name], random_locus, flip)
                                        match method:
                                            case "median":
                                                pileup_null = np.apply_along_axis(np.nanmedian, 0, random_pileup_matrices)
                                            case "mean":
                                                pileup_null = np.apply_along_axis(np.nanmean, 0, random_pileup_matrices)
                                        pileup = pileup / pileup_null
                        
                        title = f"{name.replace('_', ' ')} pileup ({len(pileup_matrices)} matrices)"
                        pileup_outpath = f"{outfolder}/{name}_pileup" if len(outfolder) > 0 else ""
                        if len(pileup_outpath) > 0:
                            if not os.path.exists(f"{outfolder}/matrices_tables"):
                                os.mkdir(f"{outfolder}/matrices_tables")
                            pd.DataFrame(pileup).to_csv(f"{outfolder}/matrices_tables/{name}_pileup.csv")
                        display_pileup(pileup, window, cmap=cmap, title=title, outpath=pileup_outpath, output_format=output_format, display_strand=flip, display_sense=display_sense)

def extract2d(cool_files, positions, outpath, params):

    if not os.path.exists(outpath):
        os.mkdir(outpath)

    # parsing parameters
    gff = params['gff']
    windows = params['windows']
    detrending = params['detrending']
    nb_pos = params['nb_pos']
    max_dist = params['random_max_dist']
    loops = params['loops']
    min_dist = params['min_dist']
    diagonal_mask = params['diagonal_mask']
    trans_contact = params['trans_contact']
    circular_chromosomes = params['circular_chromosomes']
    display_strand = params['display_strand']
    output_format = params['output_formats']
    compute_pileup = params['pileup']
    plot_loci = params['loci']
    method = params['method']
    flip = params['flip']
    cmap = params['cmap_pileup']
    display_sense = params['display_sense']
    center = params["center"]
    separate_by = params["separate_by"]
    separate_regions = params["separation_regions"]
    overlap = params["overlap"]
    contact_separation = params["contact_separation"]
    contact_range = params["contact_range"]
    ps_detrending = params["detrending"] == "ps"


    # TODO: annotate if gff is provided
    if gff != None:
        #positions_parsed = parse_bed2d_annotated(pos_file, gff, overlap=overlap)
        positions_parsed = parse_bed2d(positions, default_strand=1)
        # TODO write position file
        # write_positions(positions_parsed, outpath)
    else:
        positions_parsed = parse_bed2d(positions, default_strand=1)
    data_title = positions.split('/')[-1].split('.')[0].replace('/', '_')

    # TODO: write separate_positions_2d and delete bellow line
    selected_positions = {data_title:positions_parsed}
    # selected_positions = separate_positions_2d(positions_parsed, data_title, separate_by=separate_by, separate_regions=separate_regions, overlap=overlap, outpath=outpath)
    random_locus = []

    # TODO: apply display for 2d datas: no single loci available
    for position_name, positions in selected_positions.items():
        for cool_path in cool_files:
            cool = cooler.Cooler(cool_path)
            bins = cool.binsize
            if detrending == "patch":
                random_locus = get_random_from_locus_2d(cool, positions_parsed, nb_pos=nb_pos, max_dist=max_dist) if len(random_locus) == 0 else random_locus

            matrix_outfolder = f"{outpath}/{cool.filename.split('/')[-1].split('.')[0]}"
            if not os.path.exists(matrix_outfolder):
                os.mkdir(matrix_outfolder)


            for window in windows:
                outfolder = f"{matrix_outfolder}/window_{window//1000}kb"
                if not os.path.exists(outfolder):
                    os.mkdir(outfolder)

                # retrieving submatrices
                submatrices = compute_submatrices(cool, 
                                                  position_name, 
                                                  positions, 
                                                  bins, 
                                                  window, 
                                                  loops=loops, 
                                                  min_dist=min_dist,
                                                  circular=circular_chromosomes, 
                                                  trans_contact=trans_contact, 
                                                  diagonal_mask=diagonal_mask, 
                                                  center=center, 
                                                  sort_contact=contact_separation, 
                                                  contact_range = contact_range,
                                                  ps_detrend = ps_detrending,
                                                  is_2d=True)

                if detrending == "patch":
                    random_submatrices = compute_submatrices(cool, 
                                                            position_name, 
                                                            random_locus, 
                                                            bins, 
                                                            window, 
                                                            loops=loops, 
                                                            min_dist=min_dist,
                                                            circular=circular_chromosomes, 
                                                            trans_contact=False, 
                                                            center = center, 
                                                            sort_contact=contact_separation, 
                                                            contact_range = contact_range,
                                                            is_2d=True)

                single_positions, single_pairs = compute_pairs2d(positions)
                for name in submatrices.keys():
                    if len(submatrices[name]) == 0:
                        continue
                    # displaying submatrices
                    if plot_loci:
                        display_submatrices(submatrices[name], single_positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense)
                    display_all_submatrices(submatrices[name], single_positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense)
                    
                    # computing pileup
                    if compute_pileup:
                        ## aggregating the matrices
                        pileup_matrices = get_windows(submatrices[name], positions, flip)
                        match method:
                            case "median":
                                pileup = np.apply_along_axis(np.nanmedian, 0, pileup_matrices)
                            case "mean":
                                pileup = np.apply_along_axis(np.nanmean, 0, pileup_matrices)

                        ## detrending
                        if name[-len("trans"):] != "trans":
                            match detrending:
                                case "patch":
                                    if nb_pos >= 1:
                                        random_pileup_matrices = get_windows(random_submatrices[name], random_locus, flip)
                                        match method:
                                            case "median":
                                                pileup_null = np.apply_along_axis(np.nanmedian, 0, random_pileup_matrices)
                                            case "mean":
                                                pileup_null = np.apply_along_axis(np.nanmean, 0, random_pileup_matrices)
                                        pileup = pileup / pileup_null
                        
                        title = f"{name.replace('_', ' ')} pileup ({len(pileup_matrices)} matrices)"
                        pileup_outpath = f"{outfolder}/{name}_pileup" if len(outfolder) > 0 else ""
                        if len(pileup_outpath) > 0:
                            if not os.path.exists(f"{outfolder}/matrices_tables"):
                                os.mkdir(f"{outfolder}/matrices_tables")
                            pd.DataFrame(pileup).to_csv(f"{outfolder}/matrices_tables/{name}_pileup.csv")
                        display_pileup(pileup, window, cmap=cmap, title=title, outpath=pileup_outpath, output_format=output_format, display_strand=flip, display_sense=display_sense)


def tracks(cool_files, tracks, outpath, params):
    if not os.path.exists(outpath):
        os.mkdir(outpath)
    # parsing parameters
    threshold = params['threshold']
    percentage = params['percentage']
    gff = params['gff']
    track_unit = params['track_unit']
    overlap = params['overlap']
    binning = params['binning']
    windows = params['windows']
    detrending = params['detrending']
    nb_pos = params['nb_pos']
    max_dist = params['random_max_dist']
    loops = params['loops']
    min_dist = params['min_dist']
    diagonal_mask = params['diagonal_mask']
    trans_contact = params['trans_contact']
    circular_chromosomes = params['circular_chromosomes']
    display_strand = params['display_strand']
    output_format = params['output_formats']
    compute_pileup = params['pileup']
    plot_loci = params['loci']
    method = params['method']
    flip = params['flip']
    cmap = params['cmap_pileup']
    display_sense = params['display_sense']
    center = params["center"]
    contact_separation = params["contact_separation"]
    contact_range = params["contact_range"]
    ps_detrending = params["detrending"] == "ps"

    # parsing tracks
    tracks_parsed, bw_tracks = parse_tracks(tracks, threshold = threshold, percentage = percentage, gff=gff, binning=binning)
    data_title = tracks.split('/')[-1].split('.')[0].replace('/', '_')

    selected_positions = separate_positions(tracks_parsed, data_title, outpath=outpath)
    random_locus = []
    for position_name, positions in selected_positions.items():
        for cool_path in cool_files:
            cool = cooler.Cooler(cool_path)
            bins = cool.binsize

            if detrending == "patch":
                random_locus = get_random_from_locus(cool, tracks_parsed, nb_pos=nb_pos, max_dist=max_dist) if len(random_locus) == 0 else random_locus

            matrix_outfolder = f"{outpath}/{cool.filename.split('/')[-1].split('.')[0]}"
            if not os.path.exists(matrix_outfolder):
                os.mkdir(matrix_outfolder)


            for window in windows:
                outfolder = f"{matrix_outfolder}/window_{window//1000}kb"
                if not os.path.exists(outfolder):
                    os.mkdir(outfolder)

                # retrieving submatrices
                submatrices = compute_submatrices(cool, 
                                                  position_name, 
                                                  positions, 
                                                  bins, 
                                                  window, 
                                                  loops=loops, 
                                                  min_dist=min_dist,
                                                  circular=circular_chromosomes, 
                                                  trans_contact=trans_contact, 
                                                  diagonal_mask=diagonal_mask, 
                                                  center=center, 
                                                  sort_contact=contact_separation, 
                                                  contact_range = contact_range,
                                                  ps_detrend = ps_detrending)

                if detrending == "patch":
                    random_submatrices = compute_submatrices(cool, 
                                                            position_name, 
                                                            random_locus, 
                                                            bins, 
                                                            window, 
                                                            loops=loops, 
                                                            min_dist=min_dist,
                                                            circular=circular_chromosomes, 
                                                            trans_contact=False, 
                                                            center = center, 
                                                            sort_contact=contact_separation, 
                                                            contact_range = contact_range,
                                                            compile=True)
                    
                # computing subtracks
                subtracks = compute_subtracks(bw_tracks, positions, window, center = center, circular=circular_chromosomes)
                if detrending == "patch":
                    random_subtracks = compute_subtracks(bw_tracks, random_locus, window, center = center, circular=circular_chromosomes)
                 
                for name in submatrices.keys():
                    if len(submatrices[name]) == 0:
                        continue
                    # displaying submatrices
                    if plot_loci:
                        display_submatrices(submatrices[name], positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense, tracks=subtracks, track_label=track_unit)
                    display_all_submatrices(submatrices[name], positions, window, outfolder=outfolder + f"/{name}", circular=circular_chromosomes, chromsizes = cool.chromsizes, output_format=output_format, display_strand=display_strand, display_sense=display_sense)
                    
                    # computing pileup
                    if compute_pileup:
                        ## aggregating the matrices
                        pileup_matrices = get_windows(submatrices[name], positions, flip)
                        match method:
                            case "median":
                                pileup = np.apply_along_axis(np.nanmedian, 0, pileup_matrices)
                            case "mean":
                                pileup = np.apply_along_axis(np.nanmean, 0, pileup_matrices)

                        ## detrending
                        if name[-len("trans"):] != "trans":
                            match detrending:
                                case "patch":
                                    if nb_pos >= 1:
                                        random_pileup_matrices = get_windows(random_submatrices[name], random_locus, flip)
                                        match method:
                                            case "median":
                                                pileup_null = np.apply_along_axis(np.nanmedian, 0, random_pileup_matrices)
                                            case "mean":
                                                pileup_null = np.apply_along_axis(np.nanmean, 0, random_pileup_matrices)
                                        pileup = pileup / pileup_null
                        
                        title = f"{name.replace('_', ' ')} pileup ({len(pileup_matrices)} matrices)"
                        pileup_outpath = f"{outfolder}/{name}_pileup" if len(outfolder) > 0 else ""
                        # TODO add display strand True for separated sense of transcription
                        display_strand = False
                        if len(pileup_outpath) > 0:
                            if not os.path.exists(f"{outfolder}/matrices_tables"):
                                os.mkdir(f"{outfolder}/matrices_tables")
                            pd.DataFrame(pileup).to_csv(f"{outfolder}/matrices_tables/{name}_pileup.csv")

                        track_pileup = compile_tracks(positions, subtracks, method = method, flip = flip)
                        # TODO: put the detrending mode in the log and not the figure; only display "detrended"
                        if detrending == "patch":
                            track_pileup_null = compile_tracks(random_locus, random_subtracks, method = method, flip = flip)
                            track_pileup = track_pileup /track_pileup_null
                            track_unit = track_unit + "\n(patch detrended)"
                        if detrending == "ps":
                            mean_value = bw_tracks.header()['sumData']/bw_tracks.header()['nBasesCovered']
                            track_pileup = track_pileup / mean_value
                            track_unit = track_unit + "\n(detrended by global mean)"

                        display_strand_specified = flip and (not loops)
                        display_pileup(pileup, window, track_pileup=track_pileup, cmap=cmap, title=title, outpath=pileup_outpath, output_format=output_format, display_strand=display_strand_specified, display_sense=display_sense, is_contact=loops, track_label=f"{method}\n{track_unit}")

def annotate(gff, bed):
    pass


## compute subtracks test
# positions = positions._append({
#                     "Name":'test_up',
#                     "Chromosome":"NC_014500.1",
#                     "Start": 4920000,
#                     "End":4921000,
#                     "Strand":1,
#                     "Track":0
#                 },ignore_index = True)
#                 positions = positions._append({
#                     "Name":'test_down',
#                     "Chromosome":"NC_014500.1",
#                     "Start":12000,
#                     "End":13000,
#                     "Strand":1,
#                     "Track":0
#                 },ignore_index = True)
#                 print(positions)
#                 print(bw_tracks.chroms())

#                 for subtrack in subtracks.keys():
#                     if subtrack == 40 or subtrack == 41:
#                         print(subtracks[subtrack])


# import logging
import click

from .imports import *

from .custom_types import COOL, INT_LIST, STR_LIST, POSITION2D_FILE, GFF_FILE

import hicue.hicue as h

@click.command("extract2d")
@click.argument("outpath", type=click.Path(file_okay=False))
@click.argument('positions', type=POSITION2D_FILE)
@click.argument("cool_files", type=COOL)
@click.option('--gff', type=GFF_FILE, help="Gff file provided for the position file automatic annotation if the file is a bed2d. The positions considered for pileup are all the genes contained in the bed2d files. For more options, use the hicue annotate command.")
@click.option('-w', '--windows', type=INT_LIST, default="30000", help="Window size for sub-matrices extraction in bp. Several window sizes can be provided as a comma-separated list. Default value: 30000.")
@click.option('-d', '--detrending',type=click.Choice(['patch', 'ps', 'none'], case_sensitive=False), default='none', help='Detrending option. Default value: none.')
@click.option('-m', '--pileup_method', type=click.Choice(['median', 'mean'], case_sensitive=False), default='median', help="Aggregation method. If the selected detrending is patch, the method will also be used to aggregate the random sub-matrices. Default value: median.")
@click.option('-f', '--flip', is_flag=True, help="Enables sub-matrices flipping depending on their sense of transcription in the pileups. Requires the strand annotation of provided positions. If not provided, will consider all position in forward.")
@click.option('--nb_pos', type=int, default=2, help="Number of random positions selected for patch detrending. Default value: 2.")
@click.option('--rand_max_dist', type=int, default="100000", help="Maximum distance in bp between provided positions and the random positions selected for patch detrending. Default value: 100000.")
@click.option('--format', type=STR_LIST, default="pdf", help="Figures saving formats. Default value: pdf")
@click.option('--circulars', type=STR_LIST, default="none", help="Coma-separated list of the chromosomes to treat as circular. By default, chromosomes are not considered circular.")
@click.option('--loops', is_flag=True, help="Centers the sub-matrices on pairs of positions instead of single position.")
@click.option('--trans', is_flag=True, help="Enables trans-chromosomal contacts in extracted sub-matrices when in --loops option.")
@click.option('--min_dist', type=int, default="30000", help="Minimal distance in bp between two positions before the pair is used in the --loops option. Default value: 30000.")
@click.option('--diag_mask', type=int, default="0", help="Distance from the diagonal in bp to which the matrix are set to NaN. Is applied only if superior to the bin size. Default value: 0")
@click.option('--pileup/--no-pileup', default=True, help="Compute and display pileups.")
@click.option('--loci/--no-loci', default=False, help="Display single loci as individual figures.")
@click.option('--display_strand', is_flag=True, help="Display strands on the single matrices and pileup. Requires the strand annotation of provided positions.")
@click.option('--cmap_limits', type=(float, float), help="Min and Max value for matrix display. Usage: --cmap_limits MIN MAX.")
@click.option('--display_sense', type=click.Choice(['forward', 'reverse'], case_sensitive=False), default='forward', help="Sense of display. In 'forward' mode, the matrices are represented with the forward sense going from left to right, and from right to left in 'reverse' mode.")
@click.option('--center', type=click.Choice(['start', 'center', 'end'], case_sensitive=False), default='start', help="Defines the positional parameter of each position chosen as the window center. 'start' for the start site, 'end' for the end site, and 'center' for the average between those last two.")
@click.option('--separate_by', type=STR_LIST, default="None", help="Comma-separated list of the separation operations. Allowed operations: ['direction', 'regions', 'chroms']. As the option is an inclusion of the separate command, enter: hicue separate --help for more information. ")
@click.option('--overlap', type=click.Choice(['strict', 'flex'], case_sensitive=False), default='strict', help="When evaluating the belonging of a position to an interval, sets the severity of the discrimination: 'strict' will only allow position having start and end positions within the interval, whereas 'flex' will consider the position even if the overlap is not complete. Default value: strict.")
@click.option('--separation_regions', type=click.Path(exists=True, dir_okay=False, readable=True, path_type='csv'), help="Path to csv file providing the regions when --separate_by regions mode is selected. As the option is an inclusion of the separate command, enter: hicue separate --help for more information.")# Allows discountinuous interval if provided with the same ID. Csv format: Id,Chromosome,Start,End.")
@click.option('--contact_separation', type=click.Choice(['cis_trans','distance','none'], case_sensitive=False),  default="none", help="Separation option for the --loops.\ncis_trans: cis contacts and trans contact will be plotted separatly. Recomended parameters: --detrending ps --method_pileup mean.\ndistance: using the --contact_range parameter, will compute a selection for each distance in the range. Only applicable on cis contact (if the --trans option is enabled, will not compute the trans contacts). Default value: none.")
@click.option('--contact_range', type=(int, int, int), default=["20000", "200000", "30000"], help="Provides MIN MAX STEP in bp as a range for the distance separation on contacts. Overrides the --min_dist option. Default value: (20000,200000,30000).")
@click.pass_context
def extract2d(ctx, 
            outpath,
            positions,
            cool_files, 
            gff,
            windows, 
            detrending, 
            pileup_method,
            flip,
            nb_pos,
            rand_max_dist,
            format,
            circulars,
            loops,
            trans,
            min_dist,
            diag_mask,
            pileup,
            loci,
            display_strand,
            cmap_limits,
            display_sense,
            center, 
            separate_by,
            overlap,
            separation_regions,
            contact_separation,
            contact_range
            ):
        click.echo(f"""Extracting from {cool_files}
                positions file: {positions}
                outpath: {outpath}
                --gff = {gff}
                --windows = {windows}
                --detrending = {detrending}
                --pileup_method = {pileup_method}
                --flip = {flip}
                --nb_pos = {nb_pos}
                --rand_max_dist = {rand_max_dist}
                --format = {format}
                --circulars = {circulars}
                --loops = {loops}
                --trans = {trans}
                --min-dist = {min_dist}
                --diag_mask = {diag_mask}
                --pileup = {pileup}
                --loci = {loci}
                --display_strand = {display_strand}
                --cmap_limits = {cmap_limits}
                --display_sense = {display_sense}
                --center = {center}
                --separate_by = {separate_by}
                --overlap = {overlap}
                --separation_regions = {separation_regions}
                --contact_separation = {contact_separation}
                --contact_range = {contact_range}
        """)

        params = {
                "gff":gff,
                "windows":windows,
                "detrending":detrending,
                "nb_pos":nb_pos,
                "random_max_dist":rand_max_dist,
                "loops":loops,
                "min_dist":min_dist,
                "diagonal_mask":diag_mask,
                "trans_contact":trans,
                "circular_chromosomes":circulars,
                "display_strand":display_strand,
                "output_formats":format,
                "pileup":pileup,
                "loci":loci,
                "method":pileup_method,
                "flip":flip,
                "cmap_pileup": cmap_limits,
                "display_sense":display_sense,
                "center":center,
                "overlap":overlap,
                "separate_by":separate_by,
                "separation_regions":separation_regions,
                "contact_separation":contact_separation,
                "contact_range":contact_range
        }

        h.extract2d(cool_files, positions, outpath, params)


from .cli.imports import *

def get_random_from_locus(cool, locus, nb_pos=2, max_dist=100000):
    """From locus list, computes a list of nb_pos random locus for each loci"""
    random_loci = pd.DataFrame(columns=locus.columns)

    for i in locus.index:
        loci = locus.iloc[i]

        # TODO: include circularity in random picking
        min_pos = max(0, loci["Start"] - max_dist)
        max_pos = min(cool.chromsizes[loci["Chromosome"]], loci["End"] + max_dist)

        for _ in range(nb_pos):

            random_pos = int(np.random.random() * abs(max_pos - min_pos) + min_pos)

            loci["Start"] = random_pos
            loci["End"] = random_pos

            random_loci = random_loci._append(loci, ignore_index = True)

    return random_loci

def get_random_from_locus_2d(cool, locus, nb_pos=2, max_dist=100000):
    """From 2d locus list, computes a list of nb_pos random pair of locus for each loci pair, maintaining the distance between those."""
    random_loci = pd.DataFrame(columns=locus.columns)

    for i in locus.index:
        loci = locus.iloc[i].copy(deep=True)

        min_pos = max(0, loci["Start"] - max_dist)
        max_pos = min(cool.chromsizes[loci["Chromosome"]], loci["End"] + max_dist)

        for _ in range(nb_pos):

            random_pos = int(np.random.random() * abs(max_pos - min_pos) + min_pos)
            loci_dist = abs(random_pos - loci["Start"])

            # Two cases: 
            ## the second location is on the same chromosome, the distance is between the loci physically on the DNA strand;
            ## the second location is on another chromosome, but it is the same operation, as we try to extract the random matrices on the same diagonal as the original positions in the trans contact matrix.
            if random_pos <= loci["Start"]:
                random_pos2 = max(0, loci["Start2"] - loci_dist)
            else: 
                random_pos2 = min(cool.chromsizes[loci["Chromosome2"]], loci["Start2"] + loci_dist)


            loci["Start"] = random_pos
            loci["End"] = random_pos
            loci["Start2"] = random_pos2
            loci["End2"] = random_pos2

            random_loci = random_loci._append(loci, ignore_index = True)

    return random_loci

def detrend(matrix):
    """Applies detrending by p(s) to a chromosome's matrix"""
    y = distance_law(matrix)
    y[np.isnan(y)] = 0.0
    matrix_index = [[abs(i - j) for i in range(len(matrix))] for j in range(len(matrix))]
    matrix = matrix / y[matrix_index]
    return matrix

def extract_window(cool, locus1, locus2, binning, window, circular=[], trans=False, diagonal_mask=0, center="start", detrend_matrix = False):
    """Extracts a window from a matix given positions and parameters."""
    if trans and locus1["Chromosome"]!=locus2["Chromosome"]:
        extent1 = cool.extent(locus1["Chromosome"])
        extent2 = cool.extent(locus2["Chromosome"])
        matrix = cool.matrix(balance=True)[extent1[0]: extent1[1], extent2[0]:extent2[1]]
        if detrend_matrix:
            matrix =matrix / np.nanmean(matrix)
    else:
        matrix = detrend(cool.matrix(balance=True).fetch(locus1["Chromosome"])) if detrend_matrix else cool.matrix(balance=True).fetch(locus1["Chromosome"])

    if locus1["Chromosome"] == locus2["Chromosome"]:
        for i in range(diagonal_mask//binning):
            np.fill_diagonal(matrix[i:],  np.nan)
            np.fill_diagonal(matrix[:,i:],  np.nan)

    match center:
        case "start":
            start1 = min(locus1["Start"], locus1["End"]) if locus1["Strand"] == 1 else max(locus1["Start"], locus1["End"])
            start2 = min(locus2["Start"], locus2["End"]) if locus2["Strand"] == 1 else max(locus2["Start"], locus2["End"])
        case "center":
            start1 = (locus1["Start"] + locus1["End"]) // 2
            start2 = (locus2["Start"] + locus2["End"]) // 2
        case "end":
            start1 = max(locus1["Start"], locus1["End"]) if locus1["Strand"] == 1 else min(locus1["Start"], locus1["End"])
            start2 = max(locus2["Start"], locus2["End"]) if locus2["Strand"] == 1 else min(locus2["Start"], locus2["End"])

    pos1 = start1 // binning
    pos2 = start2 // binning
    window_binned = window // binning

    pos1_sub = np.full((window_binned * 2 + 1, len(matrix[0])), np.nan)
    if pos1 - window_binned < 0:
        if locus1["Chromosome"] in circular:
            pos1_sub = np.concatenate([
                matrix[pos1 - window_binned:],
                matrix[:pos1 + window_binned + 1]
            ], axis=0)
        else:
            pos1_sub[-(pos1 + window_binned + 1):] = matrix[:pos1 + window_binned + 1]
    elif pos1 + window_binned + 1 > len(matrix):
        if locus1["Chromosome"] in circular:
            pos1_sub = np.concatenate([
                matrix[pos1 - window_binned:],
                matrix[:pos1 + window_binned + 1 - len(matrix)]
            ], axis=0)
        else:
            pos1_sub[:-(pos1 + window_binned + 1 - len(matrix))] = matrix[pos1 - window_binned:]
    else:
        pos1_sub = matrix[pos1 - window_binned: pos1 + window_binned + 1]


    submatrix = np.full((window_binned * 2 + 1, window_binned * 2 +1), np.nan)
    if pos2 - window_binned < 0:
        if locus2["Chromosome"] in circular:
            submatrix = np.concatenate([
                pos1_sub[:, pos2 - window_binned:],
                pos1_sub[:, :pos2 + window_binned + 1]
            ], axis=1)
        else:
            submatrix[:, -(pos2 + window_binned + 1):] = pos1_sub[:, :pos2 + window_binned + 1]
    elif pos2 + window_binned + 1 > len(matrix[0]):
        if locus2["Chromosome"] in circular:
            submatrix = np.concatenate([
                pos1_sub[:, pos2 - window_binned:],
                pos1_sub[:, :pos2 + window_binned + 1 - len(matrix[0])]
            ], axis=1)
        else:
            submatrix[:, :-(pos2 + window_binned + 1 - len(matrix[0]))] = pos1_sub[:, pos2 - window_binned:]
    else:
        submatrix = pos1_sub[:, pos2 - window_binned: pos2 + window_binned + 1]
    
    return submatrix

def compute_subtracks(bw_tracks, positions, window, center = 'start', circular=[]):
    """Returns a dictionnary with each position's index as key and the corresponding track on the window."""
    chrom_sizes = bw_tracks.chroms()
    subtracks = {}
    for i, locus in positions.iterrows():
        start = locus["Start"]
        match center:
            case "start":
                start = min(locus["Start"], locus["End"]) if locus["Strand"] == 1 else max(locus["Start"], locus["End"])
            case "center":
                start = (locus["Start"] + locus["End"]) // 2
            case "end":
                start = max(locus["Start"], locus["End"]) if locus["Strand"] == 1 else min(locus["Start"], locus["End"])

        chromosome = locus["Chromosome"]
        len_chrom = chrom_sizes[chromosome]

        extracted_track = []
        if start - window < 0:
            bellow_start = abs(start - window)
            extracted_track = np.concatenate([[np.nan] * bellow_start, bw_tracks.values(chromosome, 0, start + window + 1)])
            if chromosome in circular: 
                extracted_track[:bellow_start] = bw_tracks.values(chromosome, len_chrom - bellow_start, len_chrom)

        elif start + window + 1 > len_chrom:
            up_start = start + window + 1 - len_chrom
            extracted_track = np.concatenate([bw_tracks.values(chromosome, start - window, len_chrom), [np.nan] * up_start])
            if chromosome in circular:
                extracted_track[- up_start:] = bw_tracks.values(chromosome, 0, up_start)

        else:
            extracted_track = np.array(bw_tracks.values(chromosome, start - window, start + window + 1))

        subtracks[i] = extracted_track
    return subtracks

def get_dist_positions(positions, index1, index2, center="start"):
     
    locus1 = positions.iloc[index1]
    locus2 = positions.iloc[index2]
    match center:
        case "start":
            start1 = min(locus1["Start"], locus1["End"]) if locus1["Strand"] == 1 else max(locus1["Start"], locus1["End"])
            start2 = min(locus2["Start"], locus2["End"]) if locus2["Strand"] == 1 else max(locus2["Start"], locus2["End"])
        case "center":
            start1 = (locus1["Start"] + locus1["End"]) // 2
            start2 = (locus2["Start"] + locus2["End"]) // 2
        case "end":
            start1 = max(locus1["Start"], locus1["End"]) if locus1["Strand"] == 1 else min(locus1["Start"], locus1["End"])
            start2 = max(locus2["Start"], locus2["End"]) if locus2["Strand"] == 1 else min(locus2["Start"], locus2["End"])

    if locus1["Chromosome"] == locus2["Chromosome"]:
        return abs(start1 - start2)
    else:
        return np.inf
    
def compute_pairs2d(positions):
    """Separates the 2d lines of a position table into single positions, keeping tab on the pairs indexes. Both the new table and indexation are returned."""
    tmp_positions = pd.DataFrame(columns=["Name", "Chromosome", "Start", "End", "Strand"])
    index_pairs = []
    for i, pos in positions.iterrows():
        tmp_positions = tmp_positions._append({
            "Name":pos["Name"],
            "Chromosome":pos["Chromosome"],
            "Start":pos["Start"],
            "End":pos["End"],
            "Strand":pos["Strand"]
        }, ignore_index = True)
        tmp_positions = tmp_positions._append({
            "Name":pos["Name2"],
            "Chromosome":pos["Chromosome2"],
            "Start":pos["Start2"],
            "End":pos["End2"],
            "Strand":pos["Strand2"]
        }, ignore_index = True)
        index_pairs.append((i * 2, i * 2 + 1))
    return tmp_positions, index_pairs

def compute_submatrices(cool, name, positions, binning, window, circular=[], loops = False, min_dist=0, trans_contact=False, diagonal_mask=0, center="start", sort_contact="None", contact_range="20000:100000:30000", ps_detrend = False, compile = False, is_2d=False):
    locus_pairs = {}
    tmp_locus_pairs = []
    tmp_positions = None
    # if 2d positions, the pairs are pre-computed
    if is_2d:
        tmp_positions, tmp_locus_pairs = compute_pairs2d(positions)
    else:
        tmp_positions = positions.copy()
    if loops:
        tmp_locus_pairs = np.array(list(combinations(tmp_positions.index, r=2))) if not is_2d else tmp_locus_pairs
        tmp_locus_pairs_distances = np.array([get_dist_positions(tmp_positions, i, j, center=center) for i, j in tmp_locus_pairs])
        match sort_contact:
            case "none":
                locus_pairs[name] = tmp_locus_pairs[tmp_locus_pairs_distances >= min_dist]

            case "None":
                locus_pairs[name] = tmp_locus_pairs[tmp_locus_pairs_distances >= min_dist]

            case "":
                locus_pairs[name] = tmp_locus_pairs[tmp_locus_pairs_distances >= min_dist]

            case "distance":
                contact_data = contact_range.split(':')
                contact_min = int(contact_data[0]) if len(contact_data) == 3 else 0
                contact_max = int(contact_data[1]) if len(contact_data) == 3 else 0
                contact_window = int(contact_data[2]) if len(contact_data) == 3 else 0
                tmp_locus_pairs = np.array(list(combinations(tmp_positions.index, r=2)))
                tmp_locus_pairs_distances = [get_dist_positions(tmp_positions, i, j, center=center) for i, j in tmp_locus_pairs]
                for k in range(contact_min, contact_max, contact_window):
                    min_window = k
                    max_window = k + contact_window
                    k_indexes = [i for i in range(len(tmp_locus_pairs_distances)) if min_window <= tmp_locus_pairs_distances[i] < max_window]
                    locus_pairs[f"{name}_{min_window//1000}-{max_window//1000}kb"] = tmp_locus_pairs[k_indexes]

            case "cis_trans":
                cis_list = []
                trans_list = []
                for i in range(len(tmp_positions)):
                    for j in range(i +1, len(tmp_positions)):
                        if get_dist_positions(tmp_positions, i, j, center=center) <= min_dist:
                            continue
                        if tmp_positions.iloc[i]["Chromosome"] == tmp_positions.iloc[j]["Chromosome"]:
                            cis_list.append([i, j])
                        else:
                            trans_list.append([i, j])
                locus_pairs[f"{name}_cis"] = cis_list
                locus_pairs[f"{name}_trans"] = trans_list
    else:
        locus_pairs[name] = list([(i, i) for i in tmp_positions.index]) if not is_2d else tmp_locus_pairs

    all_submatrices = {}
    for locus_name in locus_pairs.keys():

        submatrices = pd.DataFrame(columns=['Loc1', 'Loc2', 'Matrix'])
        for i, j in locus_pairs[locus_name]:
            locus1 = tmp_positions.iloc[i]
            locus2 = tmp_positions.iloc[j]
            if locus1["Chromosome"] != locus2["Chromosome"] and not trans_contact:
                continue
            submatrix= extract_window(cool, locus1, locus2, binning, window, circular = circular, trans = trans_contact, diagonal_mask=diagonal_mask, center=center, detrend_matrix = ps_detrend)
            submatrices = submatrices._append({"Loc1":i, "Loc2": j, "Matrix":submatrix.flatten()}, ignore_index=True)
        all_submatrices[locus_name] = submatrices

    return all_submatrices

def get_windows(matrices, locus, flip, fill=None):
    """Returns windows for pileup flipped if necessary. fill parameter will replace zero and nan values if provided."""
    windows = []
    for _, row in matrices.iterrows():
        i, j, matrix = row["Loc1"], row["Loc2"], row["Matrix"]
        if fill != None:
            matrix[np.isnan(matrix)] = fill
            matrix[matrix <= fill] = fill
        if i == j and flip and locus.iloc[i]['Strand'] == -1:
            windows.append(np.flip(matrix))
        else:
            windows.append(matrix)
    return np.array(windows)

def is_in_region(positions, regions, region, overlap="flex"):
    """Returns the positions from positions list present in the selected region"""
    selected_regions = regions[regions["Id"] == region]
    indexes = np.zeros(len(positions), dtype=bool)
    for i in range(len(positions)):
        position = positions.iloc[i]
        pos1 = min(position["Start"], position["End"])
        pos2 = max(position["Start"], position["End"])
        chromosome = position["Chromosome"]
        for _, reg in selected_regions[selected_regions["Chromosome"] == chromosome].iterrows():
            pos1_reg = min(reg["Start"], reg["End"])
            pos2_reg = max(reg["Start"], reg["End"])
            match overlap:
                case "flex":
                    # included if the position overlaps the regions, even not completely
                    if (pos1 >= pos1_reg and pos1 <= pos2_reg) or (pos2 >= pos1_reg and pos2 <= pos2_reg):
                        indexes[i] = True
                        break
                case "strict":
                    # included only if the position is completely in the region
                    if pos1 >= pos1_reg and pos2 <= pos2_reg:
                        indexes[i] = True
                        break
    return indexes

def sum_up_separate_by(positions_tables, positions, outpath):
    """Recapitulates the separations in a csv file"""
    with open(f"{outpath}/separate_by.csv", "w") as file:
        file.write(f"## Selected positions after separate_by ({sum([len(table) for table in positions_tables.values()])}/{len(positions)} selected)\n")
        for name in positions_tables.keys():
            file.write(f"# {name} {len(positions_tables[name])}\n")
        for name in positions_tables.keys():
            for _, position in positions_tables[name].iterrows():
                file.write(",".join([
                    name,
                    position["Name"],
                    str(position["Start"]),
                    str(position["End"]),
                    str(position["Strand"]),
                    position["Chromosome"]
                ]) + "\n")  

def separate_positions(positions, name, separate_by="", separate_regions="", overlap="flex", outpath=""):
    """Separates a table of positions according to the separate_by parameter. Allows several separations simultaneously."""
    positions_tables = {name:positions}
    sum_up = False
    for separation in separate_by:
        positions_tmp = {}
        match separation:                
            case "direct":
                for pos_name in positions_tables.keys():
                    current_positions = positions_tables[pos_name]
                    forwards = current_positions[current_positions["Strand"] == 1]
                    reverses = current_positions[current_positions["Strand"] == -1]
                    positions_tmp[f"{pos_name}_forward"] = forwards
                    positions_tmp[f"{pos_name}_reverse"] = reverses
                sum_up = True
                    
            case "regions":
                regions = pd.read_csv(separate_regions)
                if len(regions) > 0:
                    for pos_name in positions_tables.keys():
                        current_positions = positions_tables[pos_name]
                        for region in np.unique(regions["Id"]):
                            positions_tmp[f"{pos_name}_{region}"] = current_positions[is_in_region(current_positions, regions, region, overlap=overlap)]
                    sum_up = True
                else:
                    positions_tmp = positions_tables
                        

            case "chroms":
                for pos_name in positions_tables.keys():
                    current_positions = positions_tables[pos_name]
                    for chrom in np.unique(current_positions["Chromosome"]):
                        positions_tmp[f"{pos_name}_{chrom}"] = current_positions[current_positions["Chromosome"] == chrom]
                sum_up = True
            
            # contact functions are implemented in window selection
            case "cis_trans":
                positions_tmp = positions_tables
            case "distance":
                positions_tmp = positions_tables
            case "None":
                positions_tmp = positions_tables
            case "":
                positions_tmp = positions_tables

        for name in positions_tmp:
            positions_tmp[name].index = range(len(positions_tmp[name]))
        positions_tables = positions_tmp

    if len(outpath) > 0 and sum_up:
        sum_up_separate_by(positions_tables, positions, outpath) 
    return positions_tables

def separate_positions_2d(positions, name, separate_by="", separate_regions="", overlap="flex", outpath=""):
    """Separates a table of positions' pairs according to the separate_by parameter. Allows several separations simultaneously."""
    positions_tables = {name:positions}
    sum_up = False
    for separation in separate_by:
        positions_tmp = {}
        match separation:                
            case "direct": # for 2d separation if the strand has been provided, will separate in 4 categories: for/for, for/rev, rev/for, rev/rev
                for pos_name in positions_tables.keys():
                    current_positions = positions_tables[pos_name]
                    forwards = current_positions[current_positions["Strand"] == 1]
                    reverses = current_positions[current_positions["Strand"] == -1]
                    positions_tmp[f"{pos_name}_forward"] = forwards
                    positions_tmp[f"{pos_name}_reverse"] = reverses
                sum_up = True
                    
            case "regions": # will keep pairs in regions if both elements are in the region (with the overlap parameter)
                regions = pd.read_csv(separate_regions)
                if len(regions) > 0:
                    for pos_name in positions_tables.keys():
                        current_positions = positions_tables[pos_name]
                        for region in np.unique(regions["Id"]):
                            positions_tmp[f"{pos_name}_{region}"] = current_positions[is_in_region(current_positions, regions, region, overlap=overlap)]
                    sum_up = True
                else:
                    positions_tmp = positions_tables
                        

            case "chroms":
                for pos_name in positions_tables.keys():
                    current_positions = positions_tables[pos_name]
                    for chrom in np.unique(current_positions["Chromosome"]):
                        positions_tmp[f"{pos_name}_{chrom}"] = current_positions[current_positions["Chromosome"] == chrom]
                sum_up = True
            
            # contact functions are implemented in window selection
            case "cis_trans":
                positions_tmp = positions_tables
            case "distance":
                positions_tmp = positions_tables
            case "None":
                positions_tmp = positions_tables
            case "":
                positions_tmp = positions_tables

        for name in positions_tmp:
            positions_tmp[name].index = range(len(positions_tmp[name]))
        positions_tables = positions_tmp

    if len(outpath) > 0 and sum_up:
        sum_up_separate_by(positions_tables, positions, outpath) 
    return positions_tables

def compile_tracks(positions, tracks, flip = False, method="median"):
    """Compiles tracks according to selected method."""
    selected_tracks = []
    for i, pos in positions.iterrows():
        track = tracks[i]
        if flip and pos["Strand"] == -1:
            track = np.flip(track)
        selected_tracks.append(track)
    selected_tracks = np.array(selected_tracks)
    match method:
        case "median":
            track_pileup = np.nanmedian(selected_tracks, axis=0)
        case "mean":
            track_pileup = np.nanmean(selected_tracks, axis=0)

    return track_pileup
         
        


from .cli.imports import *

ARROW_LEFT = "←"
ARROW_RIGHT = "→"

def adjust_extents(ax, chrom1, chrom2, circular=[], chromsizes={}):
    extent_x = [item.get_text() for item in ax.get_xticklabels()]
    for i in range(len(extent_x)):
        if extent_x[i][0] == '−':
            extent_x[i] = str(chromsizes[chrom2]//1000 - int(extent_x[i][1:])) if chrom2 in circular else " "
        elif int(extent_x[i]) > chromsizes[chrom2]//1000:
            extent_x[i] = str(int(extent_x[i]) - chromsizes[chrom2]//1000) if chrom2 in circular else " "
    ax.set_xticklabels(extent_x)

    extent_y = [item.get_text() for item in ax.get_yticklabels()]
    for i in range(len(extent_y)):
        if extent_y[i][0] == '−':
            extent_y[i] = str(chromsizes[chrom1]//1000 - int(extent_y[i][1:])) if chrom1 in circular else " "
        elif int(extent_y[i]) > chromsizes[chrom1]//1000:
            extent_y[i] = str(int(extent_y[i]) - chromsizes[chrom1]//1000) if chrom1 in circular else " "
    ax.set_yticklabels(extent_y)

def plot_map(ax, matrix, loc1, loc2, window, locus, title="", display_sense="forward", circular = [], chromsizes={}, display_strand=False, strand_level=1.2, cmap=None, adjust=True, show_title=True):
    """Plots a single matrix on the provided axis"""
    is_contact = loc1 != loc2
    strand = not is_contact and locus.iloc[loc1]["Strand"] == -1
    pos1 = min(locus.iloc[loc1]["Start"], locus.iloc[loc1]["End"]) if locus.iloc[loc1]["Strand"] == 1 else max(locus.iloc[loc1]["Start"], locus.iloc[loc1]["End"])
    pos2 = min(locus.iloc[loc2]["Start"], locus.iloc[loc2]["End"]) if locus.iloc[loc2]["Strand"] == 1 else max(locus.iloc[loc2]["Start"], locus.iloc[loc2]["End"])

    name = f"{locus.iloc[loc1]['Name'].replace('/', '_')}" if not is_contact else f"{locus.iloc[loc1]['Name'].replace('/', '_')}-{locus.iloc[loc2]['Name'].replace('/', '_')}"
    title = name if len(title) == 0 else title
    if show_title:
        ax.set_title(title)

    log_matrix = np.log10(matrix)
    vmin = cmap[0] if cmap != None else None
    vmax = cmap[1] if cmap != None else None

    match display_sense:
        case "forward":
            mat = ax.imshow(log_matrix, extent=[(pos2 - window)//1000, (pos2 + window)//1000, (pos1 + window)//1000, (pos1 - window)//1000], cmap="afmhot_r", vmin=vmin, vmax=vmax)
        case "reverse":
            mat = ax.imshow(np.flip(log_matrix), extent=[(pos2 + window)//1000, (pos2 - window)//1000, (pos1 - window)//1000, (pos1 + window)//1000], cmap="afmhot_r", vmin=vmin, vmax=vmax)
    
    chrom1 = locus.iloc[loc1]["Chromosome"]
    chrom2 = locus.iloc[loc2]["Chromosome"]
    if adjust:
        adjust_extents(ax, chrom1, chrom2, circular, chromsizes)

    if display_strand:
        match display_sense:
            case "forward":
                transcription_sens = ARROW_LEFT if strand else ARROW_RIGHT
                arrow_alignment = "right" if strand else "left"
                pos_up = (pos1 + window * strand_level)//1000
            case "reverse":
                transcription_sens = ARROW_RIGHT if strand else ARROW_LEFT
                arrow_alignment = "left" if strand else "right"
                pos_up = (pos1 - window * strand_level)//1000
        ax.text(pos2//1000, pos_up, transcription_sens, horizontalalignment=arrow_alignment, fontsize=20)
    return mat

def display_submatrices(submatrices, locus, window, outfolder="", output_format=['pdf'], circular=[], chromsizes = {}, display_strand=False, display_sense="forward", tracks = None, track_label = ""):
    if len(outfolder) > 0 and not os.path.exists(outfolder):
        os.mkdir(outfolder)
    if len(outfolder) > 0 and not os.path.exists(outfolder + "/indiviudal_displays"):
        os.mkdir(outfolder  + "/indiviudal_displays")
    
    for _, row in submatrices.iterrows():
        i, j, matrix = row["Loc1"], row["Loc2"], row["Matrix"]
        is_contact = i!=j
        reshape_size = int(np.sqrt(len(matrix)))
        matrix = matrix.reshape((reshape_size, reshape_size))
        title = f"Window centered on\n{locus.iloc[i]['Name']} vs {locus.iloc[j]['Name']}" if is_contact else f"Window centered on\n{locus.iloc[i]['Name']}"
        y_label = f"{locus.iloc[i]['Name']}" if is_contact else ""
        x_label = f"{locus.iloc[j]['Name']}" if is_contact else ""
        outpath = "" if len(outfolder) == 0 else f"{outfolder}/indiviudal_displays/{locus.iloc[i]['Name'].replace('/', '_')}" if not is_contact else f"{outfolder}/indiviudal_displays/{locus.iloc[i]['Name'].replace('/', '_')}-{locus.iloc[j]['Name'].replace('/', '_')}"

        if tracks == None:

            plt.figure(figsize=(6,6))

            mat = plot_map(plt.gca(), matrix, i, j, window, locus, title=title, circular=circular, chromsizes = chromsizes, display_sense=display_sense, display_strand=display_strand, strand_level=1.2)

            plt.colorbar(mat, fraction=0.01)
            plot_xlabel =  "\n" + f"{locus.iloc[j]['Chromosome']} Genomic coordinates (in kb)"
            plot_xlabel += "\n" + x_label if len(x_label) > 0 else ""
            plt.xlabel(plot_xlabel)
            plot_ylabel =  f"{locus.iloc[i]['Chromosome']} Genomic coordinates (in kb)"
            plot_ylabel = y_label + "\n" + plot_ylabel if len(y_label) > 0 else plot_ylabel
            plt.ylabel(plot_ylabel)
        
        else:

            width = 3 if is_contact else 2
            wspace = 0.4 if is_contact else 0.1
            hspace = 0.7 if is_contact else 0.5
            ratios = [6, 1, 0.1] if is_contact else [6, 0.1]
            figwidth = 8 if is_contact else 6
            figheight = 7 if is_contact else 8
            plt.figure(figsize=(figwidth,figheight))
            gs = grid.GridSpec(5, width, height_ratios = [1,1,1,1,1], width_ratios = ratios, wspace=wspace, hspace=hspace) 

            # matrix ax
            ax = plt.subplot(gs[:4, 0])
            mat = plot_map(ax, matrix, i, j, window, locus, title=title, show_title=(not is_contact), circular=circular, chromsizes = chromsizes, display_sense=display_sense, display_strand=display_strand, strand_level=1.2, adjust=False)

            # colorbar ax
            ax_cb = plt.subplot(gs[1, width - 1])
            plt.colorbar(mat, fraction=0.01, cax=ax_cb)
            
            flip_tracks = display_sense == "reverse"

            # first track
            ax_track1 = plt.subplot(gs[4, 0], sharex=ax)
            track = tracks[j]
            xstart, xstop = ax.get_xlim()
            index1 = [i/len(track) * (xstop - xstart) + xstart for i in range(len(track))]
            index1 = np.flip(index1) if flip_tracks else index1
            ax_track1.plot(index1, track)
            ax_track1.set_ylabel(track_label)

            # second track
            if is_contact:
                ax_track2 = plt.subplot(gs[:4, 1],sharey = ax)
                track = tracks[i]
                xstart, xstop = ax.get_ylim()
                index2 = [i/len(track) * (xstop - xstart) + xstart for i in range(len(track))]
                index2 = index2 if flip_tracks else np.flip(index2)
                ax_track2.plot(track, index2)
                ax_track2.yaxis.tick_right()
                ax_track2.set_xlabel(track_label)

            adjust_extents(ax, locus.iloc[i]["Chromosome"], locus.iloc[j]["Chromosome"], circular, chromsizes)

            plot_xlabel =  "\n" + f"{locus.iloc[j]['Chromosome']} Genomic coordinates (in kb)"
            plot_xlabel += "\n" + x_label if len(x_label) > 0 else ""
            ax_track1.set_xlabel(plot_xlabel)
            plot_ylabel =  f"{locus.iloc[i]['Chromosome']} Genomic coordinates (in kb)"
            plot_ylabel = y_label + "\n" + plot_ylabel if len(y_label) > 0 else plot_ylabel
            ax.set_ylabel(plot_ylabel)

            if is_contact:
                plt.suptitle(title)

        if len(outpath) > 0 :
            for format in output_format:
                plt.savefig(outpath + f".{format}", bbox_inches="tight")
        else:
            plt.show()

def display_pileup(pileup, window, track_pileup=[], cmap=None, title="", outpath="", output_format=['.pdf'], display_strand=True, display_sense="forward", is_contact = False, track_label="Average Track"):
    vmin = None if cmap == None else cmap[0]
    vmax = None if cmap == None else cmap[1]
    xlabel = "\nGenomic coordinates (in kb)"
    ylabel = "Genomic coordinates (in kb)"
    pileup_sense = np.flip(pileup) if display_sense == "reverse" else pileup
    if len(pileup.shape) < 2:
        reshape_size = int(np.sqrt(len(pileup)))
        pileup_sense = pileup_sense.reshape((reshape_size, reshape_size))
    if len(track_pileup) == 0:
        plt.figure(figsize=(6,6))
        plt.title(title)
        match display_sense:
            case "forward":
                mat = plt.imshow(np.log10(pileup_sense), extent=[-window//1000, window//1000, window//1000, -window//1000], cmap="seismic", vmin=vmin, vmax=vmax)
            case "reverse":
                mat = plt.imshow(np.log10(pileup_sense), extent=[window//1000, -window//1000, -window//1000, window//1000], cmap="seismic", vmin=vmin, vmax=vmax)
        plt.colorbar(mat, fraction=0.01)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        if display_strand:
            transcription_sens = ARROW_LEFT if display_sense == "reverse" else ARROW_RIGHT
            arrow_alignment = "right" if display_sense == "reverse" else "left"
            to = -window//1000 * 1.2 if display_sense == "reverse" else window//1000 * 1.2
            plt.text(0, to, horizontalalignment=arrow_alignment, fontsize=20)

    else:
        width = 3 if is_contact else 2
        wspace = 0.4 if is_contact else 0.1
        hspace = 0.7 if is_contact else 0.5
        ratios = [6, 1, 0.1] if is_contact else [6, 0.1]
        figwidth = 8 if is_contact else 6
        figheight = 7 if is_contact else 8
        plt.figure(figsize=(figwidth,figheight))
        gs = grid.GridSpec(5, width, height_ratios = [1,1,1,1,1], width_ratios = ratios, wspace=wspace, hspace=hspace) 

        # matrix ax
        ax = plt.subplot(gs[:4, 0])
        match display_sense:
            case "forward":
                mat = plt.imshow(np.log10(pileup_sense), extent=[-window//1000, window//1000, window//1000, -window//1000], cmap="seismic", vmin=vmin, vmax=vmax)
            case "reverse":
                mat = plt.imshow(np.log10(pileup_sense), extent=[window//1000, -window//1000, -window//1000, window//1000], cmap="seismic", vmin=vmin, vmax=vmax)

        if display_strand:
            transcription_sens = ARROW_LEFT if display_sense == "reverse" else ARROW_RIGHT
            arrow_alignment = "right" if display_sense == "reverse" else "left"
            to = -window//1000 * 1.2 if display_sense == "reverse" else window//1000 * 1.2
            ax.text(0, to, transcription_sens, horizontalalignment=arrow_alignment, fontsize=20)

        # colorbar ax
        ax_cb = plt.subplot(gs[1, width - 1])
        plt.colorbar(mat, fraction=0.01, cax=ax_cb)
        
        flip_tracks = display_sense == "reverse"

        # first track
        ax_track1 = plt.subplot(gs[4, 0], sharex=ax)
        xstart, xstop = ax.get_xlim()
        index1 = [i/len(track_pileup) * (xstop - xstart) + xstart for i in range(len(track_pileup))]
        index1 = np.flip(index1) if flip_tracks else index1
        ax_track1.plot(index1, track_pileup)
        ax_track1.set_ylabel(track_label)

        # second track
        if is_contact:
            ax_track2 = plt.subplot(gs[:4, 1],sharey = ax)
            xstart, xstop = ax.get_ylim()
            index2 = [i/len(track_pileup) * (xstop - xstart) + xstart for i in range(len(track_pileup))]
            index2 = index2 if flip_tracks else np.flip(index2)
            ax_track2.plot(track_pileup, index2)
            ax_track2.yaxis.tick_right()
            ax_track2.set_xlabel(track_label)

        ax_track1.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        if is_contact:
            plt.suptitle(title)
        else:
            ax.set_title(title)

    if len(outpath) > 0 :
        for format in output_format:
            plt.savefig(outpath + f".{format}", bbox_inches="tight")
    else:
        plt.show()

def display_all_submatrices(submatrices, locus, window, outfolder="", output_format=['pdf'], circular=[], chromsizes = [], display_strand=False, display_sense="forward"):
    if len(outfolder) > 0 and not os.path.exists(outfolder):
        os.mkdir(outfolder)
    
    indexes = {}
    for index, row in submatrices.iterrows():
        i, j = row["Loc1"], row["Loc2"]
        indexes[(i,j)] = index
    keys = np.array(list(indexes.keys()))
    # displaying in batches of 64 matrices
    max_value = len(submatrices)//64 if len(submatrices) % 64 == 0 else len(submatrices)//64 + 1
    index = []
    n = 1
    for k in range(0, max_value):
        current_keys = keys[k * 64: (k+1)*64] if len(keys) > (k+1)*64 else keys[k * 64:]
        plt.figure(figsize=(20,20))
        gs = grid.GridSpec(8, 8, height_ratios=[1]*8, width_ratios=[1]*8, wspace=0.2, hspace=0.5)

        i, j = 0, 0
        for key in current_keys:
            matrix = submatrices.iloc[indexes[tuple(key)]]["Matrix"]
            reshape_size = int(np.sqrt(len(matrix)))
            matrix = matrix.reshape((reshape_size, reshape_size))
            loc1, loc2 = key
            is_contact = loc1 != loc2
            name = f"{locus.iloc[loc1]['Name'].replace('/', '_')}" if not is_contact else f"{locus.iloc[loc1]['Name'].replace('/', '_')}-{locus.iloc[loc2]['Name'].replace('/', '_')}"
            title = f"Position {n}" if not is_contact else f"Contact {n}"
            n += 1
            index.append([title, name])
            
            ax = plt.subplot(gs[j, i])
            plot_map(ax, matrix, loc1, loc2, window, locus, title=title, display_sense=display_sense,  circular=circular, chromsizes = chromsizes, display_strand=display_strand, strand_level=1.6)

            if j == 8 - 1:
                ax.set_xlabel('\nGenomic coordinates in kb')
            if i % 8 == 0:
                ax.set_ylabel('Genomic coordinates in kb\n')

            i += 1
            if i % 8 == 0:
                j += 1
                i = 0

        outpath = outfolder + f"/all_submatrix_batch_{(k + 1)}_out_of_{max_value}"
        if len(outpath) > 0 :
            for format in output_format:
                plt.savefig(outpath + f".{format}", bbox_inches="tight")
        else:
            plt.show()
    
    if len(outpath) > 0 :
        pd.DataFrame(index, columns=['Reference', 'Name']).to_csv(outfolder + "/all_submatrix_references.csv")
            


import numpy as np

import cooler
from BCBio import GFF
import pandas as pd
pd.options.mode.chained_assignment = None 
import numpy as np
np.seterr(all="ignore")
from itertools import combinations
import matplotlib.pyplot as plt
from matplotlib import gridspec as grid
from chromosight.utils.preprocessing import distance_law
import pyBigWig

import os
